// ignore_for_file: non_constant_identifier_names, empty_catches

import 'dart:async';
import 'package:galaxeus_lib/galaxeus_lib.dart';
import "package:telegram_client/telegram_client/extension/extension.dart";
import 'package:telegram_client/tdlib/tdlib_ffi/tdlib.dart';
import 'package:telegram_client/telegram_bot_api/update_bot.dart';
import 'package:telegram_client/telegram_client.dart';
import "package:telegram_client/telegram_client/scheme/scheme.dart"
    as telegram_client_scheme;

/// return original data json
class TelegramClient {
  late Tdlib tdlib;
  late TelegramBotApi telegramBotApi;
  EventEmitter event_emitter = EventEmitter();
  String event_update;
  String event_invoke;

  /// return original data json
  TelegramClient({
    EventEmitter? eventEmitter,
    this.event_invoke = "tg_client_invoke",
    this.event_update = "tg_client_update",
  }) {
    if (eventEmitter != null) {
      event_emitter = eventEmitter;
    }
  }

  /// return original data json
  Future<void> init({
    String? pathTdlib,
    TelegramClientTdlibOption? telegramClientTdlibOption,
    TelegramClientTelegramBotApiOption? telegramClientTelegramBotApiOption,
  }) async {
    telegramClientTdlibOption ??= TelegramClientTdlibOption(
      isAutoGetChat: false,
      is_cli: false,
      timeOutUpdate: 1.0,
      delayInvoke: Duration(milliseconds: 1),
      delayUpdate: Duration(milliseconds: 1),
      isInvokeThrowOnError: false,
    );
    telegramClientTelegramBotApiOption ??= TelegramClientTelegramBotApiOption(
      clientOption: {},
    );
    tdlib = Tdlib(
      pathTdl: pathTdlib,
      is_cli: telegramClientTdlibOption.is_cli,
      clientOption: telegramClientTdlibOption.clientOption,
      clientId: telegramClientTdlibOption.clientId,
      invokeTimeOut: telegramClientTdlibOption.invokeTimeOut,
      event_invoke: event_invoke,
      event_update: event_update,
      delayUpdate: telegramClientTdlibOption.delayUpdate,
      delayInvoke: telegramClientTdlibOption.delayInvoke,
      isAutoGetChat: telegramClientTdlibOption.isAutoGetChat,
      on_generate_extra_invoke:
          telegramClientTdlibOption.on_generate_extra_invoke,
      on_get_invoke_data: telegramClientTdlibOption.on_get_invoke_data,
      on_receive_update: telegramClientTdlibOption.on_receive_update,
      isInvokeThrowOnError: telegramClientTdlibOption.isInvokeThrowOnError,
      eventEmitter: event_emitter,
    );
    telegramBotApi = TelegramBotApi(
      tokenBot: "",
      event_invoke: event_invoke,
      event_update: event_update,
      clientOption: telegramClientTelegramBotApiOption.clientOption,
      eventEmitter: event_emitter,
      alfred: telegramClientTelegramBotApiOption.alfred,
      telegramCryptoKey: telegramClientTelegramBotApiOption.telegramCryptoKey,
      telegramUrlWebhook: telegramClientTelegramBotApiOption.telegramUrlWebhook,
    );
  }

  /// return original data json
  Listener on({
    required String event_name,
    required FutureOr<dynamic> Function(
            UpdateTelegramClient updateTelegramClient)
        onUpdate,
    required FutureOr<dynamic> Function(Object error, StackTrace stackTrace)
        onError,
  }) {
    return event_emitter.on(event_name, null, (ev, context) async {
      try {
        if (ev.eventData is UpdateBot) {
          UpdateBot updateBot = (ev.eventData as UpdateBot);
          await onUpdate(
            UpdateTelegramClient(
              rawData: updateBot.body,
              query: updateBot.query,
              client_option: {},
              tg: this,
              telegramClientData: TelegramClientData(
                tdlib_client_id: 0,
                telegram_bot_api_token_bot: "",
                telegramClientType: TelegramClientType.telegam_bot_api,
              ),
            ),
          );
        }

        if (ev.eventData is TdlibIsolateReceiveData) {
          TdlibIsolateReceiveData tdlibIsolateReceiveData =
              (ev.eventData as TdlibIsolateReceiveData);

          UpdateTd updateTd = UpdateTd(
            update: tdlibIsolateReceiveData.updateData,
            client_id: tdlibIsolateReceiveData.clientId,
            client_option: () {
              try {
                TdlibClient? tdlibClient =
                    tdlib.clients[tdlibIsolateReceiveData.clientId];
                if (tdlibClient != null) {
                  return tdlibClient.client_option;
                }
              } catch (e) {}
              return {};
            }(),
          );
          await onUpdate(
            UpdateTelegramClient(
              rawData: updateTd.raw,
              query: {},
              tg: this,
              client_option: updateTd.client_option,
              telegramClientData: TelegramClientData(
                tdlib_client_id: updateTd.client_id,
                telegram_bot_api_token_bot: "",
                telegramClientType: TelegramClientType.tdlib,
              ),
            ),
          );
        }
      } catch (e, stack) {
        try {
          await onError(e, stack);
        } catch (e) {}
      }
    });
  }

  /// return original data json
  FutureOr<telegram_client_scheme.JsonScheme> invoke({
    required Map parameters,
    required TelegramClientData telegramClientData,
    bool is_form = false,
    String? urlApi,
    String? clientType,
    void Function(int bytes, int bytess)? onUploadProgress,
    bool isVoid = false,
    Duration? delayDuration,
    Duration? invokeTimeOut,
    String? extra,
    bool? isAutoGetChat,
    bool isInvokeThrowOnError = true,
    FutureOr<String> Function(int client_id, LibTdJson libTdJson)?
        onGenerateExtraInvoke,
    FutureOr<Map<dynamic, dynamic>> Function(
            String, int client_id, LibTdJson libTdJson)?
        onGetInvokeData,
  }) async {
    // telegramClientLib ??= telegram_client_lib;
    if (telegramClientData.telegramClientType ==
        TelegramClientType.telegam_bot_api) {
      Map respond = await telegramBotApi.invoke(
        parameters["@type"],
        tokenBot: telegramClientData.telegram_bot_api_token_bot,
        parameters: parameters,
        is_form: is_form,
        urlApi: urlApi,
        clientType: clientType,
        isThrowOnError: isInvokeThrowOnError,
        onUploadProgress: onUploadProgress,
      );
      if (respond["@type"] == "error") {
        return telegram_client_scheme.Error(respond);
      }
      return telegram_client_scheme.JsonScheme(respond);
    }
    if ((telegramClientData.telegramClientType == TelegramClientType.tdlib)) {
      Map respond = (await tdlib.invoke(
        parameters["@type"],
        parameters: parameters,
        clientId: telegramClientData.tdlib_client_id,
        isVoid: isVoid,
        delayDuration: delayDuration,
        invokeTimeOut: invokeTimeOut,
        extra: extra,
        isAutoGetChat: isAutoGetChat,
        isInvokeThrowOnError: isInvokeThrowOnError,
        onGenerateExtraInvoke: onGenerateExtraInvoke,
        onGetInvokeData: onGetInvokeData,
      ));
      if (respond["@type"] == "error") {
        return telegram_client_scheme.Error(respond);
      }
      return telegram_client_scheme.JsonScheme(respond);
    }
    telegram_client_scheme.Error error = telegram_client_scheme.Error.create(
      special_extra: "error",
      code: 500,
      message: "telegram_client_type_not_found",
      description: "",
    );
    if (isInvokeThrowOnError == true) {
      throw error;
    }
    return error;
  }

  /// return original data json
  FutureOr<telegram_client_scheme.JsonScheme> request({
    required Map parameters,
    required TelegramClientData telegramClientData,
    bool is_form = false,
    String? urlApi,
    String? clientType,
    void Function(int bytes, int bytess)? onUploadProgress,
    bool isVoid = false,
    Duration? delayDuration,
    Duration? invokeTimeOut,
    String? extra,
    bool? isAutoGetChat,
    bool isInvokeThrowOnError = true,
    FutureOr<String> Function(int client_id, LibTdJson libTdJson)?
        onGenerateExtraInvoke,
    FutureOr<Map<dynamic, dynamic>> Function(
            String, int client_id, LibTdJson libTdJson)?
        onGetInvokeData,
  }) async {
    if (telegramClientData.telegramClientType != TelegramClientType.tdlib) {
      return await invoke(
        parameters: parameters,
        telegramClientData: telegramClientData,
        is_form: is_form,
        urlApi: urlApi,
        clientType: clientType,
        onUploadProgress: onUploadProgress,
        isVoid: isVoid,
        delayDuration: delayDuration,
        invokeTimeOut: invokeTimeOut,
        extra: extra,
        isAutoGetChat: isAutoGetChat,
        isInvokeThrowOnError: isInvokeThrowOnError,
        onGenerateExtraInvoke: onGenerateExtraInvoke,
        onGetInvokeData: onGetInvokeData,
      );
    }
    String method = (parameters["@type"] as String);

    /// return original data json
    FutureOr<telegram_client_scheme.JsonScheme> callApiInvoke({
      required Map parameters,
    }) async {
      return await invoke(
        parameters: parameters,
        telegramClientData: telegramClientData,
        is_form: is_form,
        urlApi: urlApi,
        clientType: clientType,
        onUploadProgress: onUploadProgress,
        isVoid: isVoid,
        delayDuration: delayDuration,
        invokeTimeOut: invokeTimeOut,
        extra: extra,
        isAutoGetChat: isAutoGetChat,
        isInvokeThrowOnError: isInvokeThrowOnError,
        onGenerateExtraInvoke: onGenerateExtraInvoke,
        onGetInvokeData: onGetInvokeData,
      );
    }

    if (RegExp(r"^(getChat)$", caseSensitive: false).hasMatch(method)) {
      return await getChat(
        parameters: parameters,
        callApiInvoke: callApiInvoke,
      );
    }
    if (RegExp(r"^(getUser)$", caseSensitive: false).hasMatch(method)) {
      return await getUser(
        parameters: parameters,
        callApiInvoke: callApiInvoke,
      );
    }

    telegram_client_scheme.Error error = telegram_client_scheme.Error.create(
      special_extra: "error",
      code: 500,
      message: "telegram_client_type_not_found",
      description: "",
    );
    if (isInvokeThrowOnError == true) {
      throw error;
    }
    return error;
  }
}
